from logging import getLogger

from .utils import core

logger = getLogger(__name__)


import numpy as np
import tensorflow as tf
from .utils.prioritised_experience_replay import PrioritizedReplayBuffer


class ReplayBuffer:
    """
    A simple FIFO experience replay buffer for NAF_debug agents.
    """

    def __init__(self, obs_dim, act_dim, size):
        self.obs1_buf = np.zeros([size, obs_dim], dtype=np.float32)
        self.obs2_buf = np.zeros([size, obs_dim], dtype=np.float32)
        self.acts_buf = np.zeros([size, act_dim], dtype=np.float32)
        self.rews_buf = np.zeros(size, dtype=np.float32)
        self.done_buf = np.zeros(size, dtype=np.float32)
        self.ptr, self.size, self.max_size = 0, 0, size

    def store(self, obs, act, rew, next_obs, done):
        self.obs1_buf[self.ptr] = obs
        self.obs2_buf[self.ptr] = next_obs
        self.acts_buf[self.ptr] = act
        self.rews_buf[self.ptr] = rew
        self.done_buf[self.ptr] = done
        self.ptr = (self.ptr + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def sample_batch(self, batch_size=32):
        if self.size < batch_size:
            idxs = np.arange(self.size)
        else:
            idxs = np.random.randint(0, self.size, size=batch_size)
        return dict(obs1=self.obs1_buf[idxs],
                    obs2=self.obs2_buf[idxs],
                    acts=self.acts_buf[idxs],
                    rews=self.rews_buf[idxs],
                    done=self.done_buf[idxs])


class ReplayBufferPER(PrioritizedReplayBuffer):
    """
    A simple FIFO experience replay buffer for NAF_debug agents.
    """

    def __init__(self, obs_dim, act_dim, size, prio_info):
        self.alpha = prio_info.get('alpha')
        self.beta = prio_info.get('beta')
        super(ReplayBufferPER, self).__init__(size, self.alpha)
        self.ptr, self.size, self.max_size = 0, 0, size

    def store(self, obs, act, rew, next_obs, done):
        super(ReplayBufferPER, self).add(obs, act, rew, next_obs, done, 1)
        self.ptr = (self.ptr + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def sample_batch(self, batch_size=32):
        if self.size < batch_size:
            batch_size = self.size
            obs1, acts, rews, obs2, done, gammas, weights, idxs = super(ReplayBufferPER, self).sample_normal(batch_size)
        else:
            obs1, acts, rews, obs2, done, gammas, weights, idxs = super(ReplayBufferPER, self).sample(batch_size,
                                                                                                      self.beta)
        return dict(obs1=obs1,
                    obs2=obs2,
                    acts=acts,
                    rews=rews,
                    done=done), [weights, idxs]


class NAF(object):
    def __init__(self, sess,
                 env, stat,
                 discount, batch_size, learning_rate,
                 max_steps, update_repeat, max_episodes, tau, pretune = None, prio_info=dict(), **nafnet_kwargs):

        self.pretune = pretune
        self.prio_info = prio_info
        self.per_flag = bool(self.prio_info)
        print('PER is:', self.per_flag)
        self.sess = sess
        self.env = env

        self.x_ph, self.a_ph, self.mu, self.V, self.Q, self.P, self.A, self.vars_pred \
            = core.mlp_normalized_advantage_function(env.observation_space.shape, act_dim=env.action_space.shape,
                                                     **nafnet_kwargs,
                                                     scope='main')
        self.x_ph_targ, self.a_ph_targ, self.mu_targ, self.V_targ, self.Q_targ, self.P_targ, self.A_targ, \
        self.vars_targ \
            = core.mlp_normalized_advantage_function(env.observation_space.shape, act_dim=env.action_space.shape,
                                                     **nafnet_kwargs,
                                                     scope='target')
        self.stat = stat
        self.discount = discount
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.action_size = env.action_space.shape[0]
        self.obs_dim = env.observation_space.shape[0]

        self.max_steps = max_steps
        self.update_repeat = update_repeat
        self.max_episodes = max_episodes

        if not (self.per_flag):
            self.replay_buffer = ReplayBuffer(obs_dim=self.obs_dim, act_dim=self.action_size, size=int(1e6))
        else:
            self.replay_buffer = ReplayBufferPER(obs_dim=self.obs_dim, act_dim=self.action_size, size=int(1e6),
                                                 prio_info=prio_info)

        with tf.name_scope('optimizer'):
            self.target_y = tf.placeholder(tf.float32, [None], name='target_y')
            self.per_weights = tf.placeholder(tf.float32, [None], name='per_weights')
            # self.loss = tf.reduce_mean(tf.squared_difference(self.target_y, tf.squeeze(self.Q)),
            #                            name='loss')
            self.loss = tf.losses.mean_squared_error(self.target_y, tf.squeeze(self.Q), weights=self.per_weights)
            self.optim = tf.train.AdamOptimizer(self.learning_rate).minimize(self.loss)
            # self.gvs = self.optim.compute_gradients(self.loss)
            # self.modified_gvs = [(tf.clip_by_value(grad, -1., 1.), var) for grad, var in self.gvs]
            # self.train_op = self.optim.apply_gradients(self.modified_gvs)

        self.target_init = tf.group([tf.assign(v_targ, v_main)
                                     for v_main, v_targ in zip(self.vars_pred, self.vars_targ)])

        # Polyak averaging for target variables (previous soft update)
        polyak = 1 - tau
        self.target_update = tf.group([tf.assign(v_targ, polyak*v_targ + (1 - polyak) * v_main)
                                       for v_main, v_targ in zip(self.vars_pred, self.vars_targ)])
        self.losses = []
        self.vs = []

    def run(self, is_train=True):
        print('Training:', is_train)
        # tf.initialize_all_variables().run()
        self.stat.set_variables(self.vars_pred)
        self.stat.load_model()  # including init
        # if is_train:
        #     self.sess.run(self.target_init)

        # pretune------------------------------------------------------------------------------
        if not(self.pretune is None):
            scan_data = self.pretune
            print('Length of scan data is: ', len(scan_data))

            if scan_data:
                for i, data in enumerate(scan_data):
                    o, a, r, o2, d, _ = data
                    self.replay_buffer.store(o, a, r, o2, d)
                    print("Number: ", i)
                    print(o, a, r, o2, d)

            batch_size_temp = self.batch_size
            self.batch_size = 10
            for _ in range(10*len(scan_data)):
                q, v, a, l = self.perceive()
                if self.stat:
                    self.stat.on_step(a, r, d, q, v, a, l)

            self.batch_size = batch_size_temp
        # -------------------------------------------------------------------------

        for self.idx_episode in range(self.max_episodes):
            o = self.env.reset()
            for t in range(0, self.max_steps):
                # 1. predict
                a = self.predict(o, is_train)
                # 2. step
                o2, r, d, _ = self.env.step(a)
                if is_train:
                    self.replay_buffer.store(o, a, r, o2, d)
                o = o2
                d = False if t == self.max_steps - 1 else d
                # 3. perceive
                if is_train:
                    pass
                    q, v, a, l = self.perceive()
                    if self.stat:
                        self.stat.on_step(a, r, d, q, v, a, l)
                if d:
                    break

    def predict(self, state, is_train):
        u = self.sess.run(self.mu, feed_dict={self.x_ph: [state]})[0]
        if is_train:
            noise_scale = 1 / (self.idx_episode + 1)
            return u + noise_scale * np.random.randn(self.action_size)
        else:
            return u

    def perceive(self):
        q_list = []
        v_list = []
        a_list = []
        l_list = []

        for iteration in range(self.update_repeat):
            if self.per_flag:
                batch, priority_info = self.replay_buffer.sample_batch(self.batch_size)
            else:
                batch = self.replay_buffer.sample_batch(self.batch_size)

            o = batch['obs1']
            o2 = batch['obs2']
            a = batch['acts']
            r = batch['rews']
            w = priority_info[0]

            v = self.sess.run(self.V_targ, feed_dict={self.x_ph_targ: o2, self.a_ph_targ: a})
            target_y = self.discount * np.squeeze(v) + r

            _, l, q, v, a = self.sess.run([
                self.optim, self.loss,
                self.Q, self.V, self.A,
            ], {
                self.target_y: target_y,
                self.x_ph: o,
                self.a_ph: a,
                self.per_weights: w
            })

            q_list.extend(q)
            v_list.extend(v)
            a_list.extend(a)
            l_list.append(l)

            # self.target_network.soft_update_from(self.pred_network)

            if self.per_flag:
                priorities = np.ones(priority_info[0].shape[-1]) * (abs(l) * 1 + 1e-7)
                self.replay_buffer.update_priorities(idxes=priority_info[1], priorities=priorities)

            logger.debug("q: %s, v: %s, a: %s, l: %s" \
                         % (np.mean(q), np.mean(v), np.mean(a), np.mean(l)))
        self.sess.run(self.target_update)
        self.losses.append(np.mean(l))
        self.vs.append(np.mean(v))

        return np.sum(q_list), np.sum(v_list), np.sum(a_list), np.sum(l_list)



#! -*- coding:utf-8 -*-

import math


class SumTree(object):
    def __init__(self, max_size):
        self.max_size = max_size
        self.tree_level = math.ceil(math.log(max_size + 1, 2)) + 1
        self.tree_size = 2 ** self.tree_level - 1
        self.tree = [0 for i in range(self.tree_size)]
        self.data = [None for i in range(self.max_size)]
        self.size = 0
        self.cursor = 0

    def add(self, contents, value):
        index = self.cursor
        self.cursor = (self.cursor + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

        self.data[index] = contents
        self.val_update(index, value)

    def get_val(self, index):
        tree_index = 2 ** (self.tree_level - 1) - 1 + index
        return self.tree[tree_index]

    def val_update(self, index, value):
        tree_index = 2 ** (self.tree_level - 1) - 1 + index
        diff = value - self.tree[tree_index]
        self.reconstruct(tree_index, diff)

    def reconstruct(self, tindex, diff):
        self.tree[tindex] += diff
        if not tindex == 0:
            tindex = int((tindex - 1) / 2)
            self.reconstruct(tindex, diff)

    def find(self, value, norm=True):
        if norm:
            value *= self.tree[0]
        return self._find(value, 0)

    def _find(self, value, index):
        if 2 ** (self.tree_level - 1) - 1 <= index:
            return self.data[index - (2 ** (self.tree_level - 1) - 1)], self.tree[index], index - (
                        2 ** (self.tree_level - 1) - 1)

        left = self.tree[2 * index + 1]

        if value <= left:
            return self._find(value, 2 * index + 1)
        else:
            return self._find(value - left, 2 * (index + 1))

    def print_tree(self):
        for k in range(1, self.tree_level + 1):
            for j in range(2 ** (k - 1) - 1, 2 ** k - 1):
                print(self.tree[j], end=' ')
            print()

    def filled_size(self):
        return self.size


if __name__ == '__main__':
    s = SumTree(10)
    for i in range(20):
        s.add(2 ** i, i)
    s.print_tree()
    print(s.find(0.5))


import os
import time

import numpy as np
import tensorflow as tf
from logging import getLogger

logger = getLogger(__name__)


class Statistic(object):
    def __init__(self, sess, env_name, model_dir, max_update_per_step, is_continued=True, save_frequency=25):

        self.save_frequency = save_frequency
        self.is_continued = is_continued
        self.sess = sess
        self.env_name = env_name
        self.max_update_per_step = max_update_per_step
        self.average_rewards = []
        self.counter = -1
        self.reset()
        self.max_avg_r = None
        self.var_list = None

        with tf.variable_scope('t'):
            self.t_op = tf.Variable(0, trainable=False, name='t')
            self.t_add_op = self.t_op.assign_add(1)

        self.model_dir = model_dir
        self.init_logging()
        self.saver = tf.train.Saver()

        with tf.variable_scope('summary'):
            scalar_summary_tags = ['total r', 'avg r', 'avg q', 'avg v', 'avg a', 'avg l']

            self.summary_placeholders = {}
            self.summary_ops = {}

            for tag in scalar_summary_tags:
                self.summary_placeholders[tag] = tf.placeholder('float32', None, name=tag.replace(' ', '_'))
                self.summary_ops[tag] = tf.summary.scalar('%s/%s' % (self.env_name, tag),
                                                          self.summary_placeholders[tag])

    def init_logging(self):
        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir)
        elif not (self.is_continued):
            for f in os.listdir(self.model_dir):
                print('Deleting: ', self.model_dir + '/' + f)
                os.remove(self.model_dir + '/' + f)
            time.sleep(.5)

    def set_variables(self, var_list):
        # self.var_list = var_list +
        self.saver = tf.compat.v1.train.Saver(self.var_list)

    def reset(self):
        self.total_q = 0.
        self.total_v = 0.
        self.total_a = 0.
        self.total_l = 0.
        self.ep_step = 0
        self.ep_rewards = []

    def on_step(self, action, reward, terminal, q, v, a, l):
        self.counter += 1
        self.t = self.t_add_op.eval(session=self.sess)

        self.total_q += q
        self.total_v += v
        self.total_a += a
        self.total_l += l

        self.ep_step += 1
        self.ep_rewards.append(reward)

        if self.counter % self.save_frequency == 0:
            print('saving...')
            self.save_model(self.t)

        if terminal:
            avg_q = self.total_q / self.ep_step / self.max_update_per_step
            avg_v = self.total_v / self.ep_step / self.max_update_per_step
            avg_a = self.total_a / self.ep_step / self.max_update_per_step
            avg_l = self.total_l / self.ep_step / self.max_update_per_step

            avg_r = np.mean(self.ep_rewards)
            total_r = np.sum(self.ep_rewards)

            print('t: %d, R: %.3f, r: %.3f, q: %.3f, v: %.3f, a: %.3f, l: %.3f' \
                  % (self.t, total_r, avg_r, avg_q, avg_q, avg_a, avg_l))
            self.average_rewards.append(avg_r)

            if self.max_avg_r == None:
                self.max_avg_r = avg_r

            # if self.max_avg_r * 0.9 <= avg_r:
            #   print('saving...')
            #   self.save_model(self.t)
            #   self.max_avg_r = max(self.max_avg_r, avg_r)

            self.inject_summary({
                'total r': total_r, 'avg r': avg_r,
                'avg q': avg_q, 'avg v': avg_v, 'avg a': avg_a, 'avg l': avg_l,
            }, self.t)

            self.reset()

    def inject_summary(self, tag_dict, t):
        summary_str_lists = self.sess.run([self.summary_ops[tag] for tag in tag_dict.keys()], {
            self.summary_placeholders[tag]: value for tag, value in tag_dict.items()
        })

    def save_model(self, t):
        # print("Saving checkpoints... implement this again")
        model_name = type(self).__name__
        folder_name = self.model_dir
        self.saver.save(self.sess, folder_name + "test.ckpt", global_step=self.t)

    def load_model(self):
        logger.info("Loading checkpoints...")
        # initialize all of the variables
        init_op = tf.global_variables_initializer()  # create the graph

        self.sess.run(init_op)
        # [print(x) for x in tf.global_variables()]

        folder_name = self.model_dir
        ckpt = tf.train.get_checkpoint_state(folder_name)
        print(ckpt)
        if ckpt and ckpt.model_checkpoint_path:
            ckpt_name = os.path.basename(ckpt.model_checkpoint_path)
            fname = os.path.join(self.model_dir, ckpt_name)
            print(fname)
            # val = [x for x in tf.global_variables() if 'target/V/fully_connected/weights:0' in x.name][0]
            # print(self.sess.run(val))

            self.saver.restore(self.sess, fname)
            # [print(x) for x in tf.global_variables()]

            logger.info("Load SUCCESS: ")
            # val = [x for x in tf.global_variables() if 'target/V/fully_connected/weights:0' in x.name][0]
            # print(self.sess.run(val))
            print(20 * '* *')
        else:
            logger.info("Load FAILED: ")
            print(20 * ' ---')
        self.t = self.t_add_op.eval(session=self.sess)



import os

from setuptools import setup, find_packages

try:
    from pypandoc import convert, convert_text


    def read_md(file_path):
        return convert_text(file_path, to='rst', format='markdown')

except ImportError:
    convert = None
    print(
        "warning: pypandoc module not found, "
        "could not convert Asciidoc to RST"
    )

    def read_md(file_path):
         with open(file_path, 'r') as f:
            return f.read()

README = os.path.join(os.path.dirname(__file__), 'README.md')


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def reqs(*f):
    return list(filter(None, [strip_comments(l) for l in open(
        os.path.join(os.getcwd(), *f)).readlines()]))

setup(
    name='pernaf',
    version="0.0.12",
    author="Simon Hirlaender",
    author_email="simon.hirlaender@cern.ch",
    description="An implementation of the Normalized Advantage Function Reinforcement Learning Algorithm with "
                "Prioritized Experience Replay",
    url = 'https://test.pypi.org/project/per-naf/',
    long_description=read_md(README),
    long_description_content_type='text/x-rst',
    install_requires=reqs('requirements.txt'),
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)


import numpy as np
import tensorflow as tf
from tensorflow.contrib.framework import get_variables
from tensorflow.contrib.layers import fully_connected


def placeholder(dim=None):
    return tf.placeholder(dtype=tf.float32, shape=(None, dim) if dim else (None,))


def placeholders(*args):
    return [placeholder(dim) for dim in args]


def fc(layer, num_outputs, scope='fc'):
    batch_norm_args = {}
    with tf.variable_scope(scope):
        return fully_connected(
            layer,
            num_outputs=int(num_outputs),
            activation_fn=tf.tanh,
            weights_initializer=tf.random_uniform_initializer(-0.05, 0.05),
            weights_regularizer=None,
            biases_initializer=tf.constant_initializer(0.0),
            # scope=scope,
            ** batch_norm_args
        )

def get_vars(scope):
    return [x for x in tf.global_variables() if scope in x.name]

def count_vars(scope):
    v = get_vars(scope)
    return sum([np.prod(var.shape.as_list()) for var in v])

"""
Normalize Advantage Function
"""


def mlp_normalized_advantage_function(x, act_dim, hidden_sizes=(100,100), activation=tf.tanh,
                                      output_activation=tf.tanh, action_space=None, weight_init=None,
                                      act_multiplier=1, scope=None):
    with tf.variable_scope(scope):
        # act_dim = a.shape.as_list()[-1]
        act_dim = act_dim[0]
        # act_limit = action_space.high[0] * act_multiplier

        x_ph = tf.placeholder(tf.float32, (None,) + tuple(x), name='observations')
        a_ph = tf.placeholder(tf.float32, (None, act_dim), name='actions')

        # create a shared network for the variables
        with tf.name_scope('hidden'):
            h = x_ph
            for idx, hidden_dim in enumerate(hidden_sizes):
                h = fc(h, hidden_dim, scope='hid%d' % idx)

        with tf.name_scope('value'):
            V = fc(h, 1, scope='V')

        with tf.name_scope('advantage'):
            l = fc(h, (act_dim * (act_dim + 1) / 2), scope='l')
            mu = fc(h, act_dim, scope='mu')

            pivot = 0
            rows = []
            for idx in range(act_dim):
                count = act_dim - idx
                diag_elem = tf.exp(tf.slice(l, (0, pivot), (-1, 1)))
                non_diag_elems = tf.slice(l, (0, pivot + 1), (-1, count - 1))
                row = tf.pad(tf.concat((diag_elem, non_diag_elems), 1), ((0, 0), (idx, 0)))
                rows.append(row)
                pivot += count
            L = tf.transpose(tf.stack(rows, axis=1), (0, 2, 1))
            P = tf.matmul(L, tf.transpose(L, (0, 2, 1)))
            tmp = tf.expand_dims(a_ph - mu, -1)
            A = -tf.matmul(tf.transpose(tmp, [0, 2, 1]), tf.matmul(P, tmp)) / 2
            A = tf.reshape(A, [-1, 1])

        with tf.name_scope('Q'):
            Q = A + V

            # print(mu.name, V.name, Q.name, P.name, A.name, h.name)
        vars = get_variables(scope)
        return x_ph, a_ph, mu, V, Q, P, A, vars


if __name__ == '__main__':
    import gym
    import matplotlib.pyplot as plt

    tf.set_random_seed(123)
    np.random.seed(123)

    MAX_POS = 1
    action_space = gym.spaces.Box(low=-MAX_POS, high=MAX_POS, shape=(1,),
                                  dtype=np.float32)
    observation_space = gym.spaces.Box(low=-MAX_POS, high=MAX_POS, shape=(1,),
                                       dtype=np.float32)

    obs_dim = observation_space.shape[0]
    act_dim = action_space.shape[0]

    nafnet_kwargs = dict(hidden_sizes=[1, 1], activation=tf.tanh,
                         weight_init=tf.random_uniform_initializer(-0.05, 0.05))
    # Share information about action space with pernaf architecture
    nafnet_kwargs['action_space'] = action_space

    # Inputs to computation graph
    # x_ph, a_ph, x2_ph, r_ph, d_ph = placeholders(obs_dim, act_dim, obs_dim, None, None)
    x_ph, a_ph = placeholders(obs_dim, act_dim)
    x2_ph, r_ph, d_ph = placeholders(obs_dim, None, None)
    # Main outputs from computation graph
    with tf.variable_scope('main'):

        print('generate main network')
        mu_pred, V_pred, Q_pred, P_pred, A_pred = \
            mlp_normalized_advantage_function(x_ph, a_ph, **nafnet_kwargs)
    # Target networks
    with tf.variable_scope('target'):
        print('generate target network')
        _, V_targ, _, _, _ = mlp_normalized_advantage_function(x2_ph, a_ph, **nafnet_kwargs)

    sess = tf.Session()
    sess.run(tf.global_variables_initializer())

    test_vector = [[i] for i in range(-100, 100)]
    a = sess.run([mu_pred, V_pred, Q_pred, P_pred, A_pred], feed_dict={x_ph: test_vector,
                                     a_ph: test_vector})
    plt.plot(np.squeeze(a[0]))
    plt.plot(np.squeeze(a[1]))
    plt.title('Mu_pred - me')
    plt.show()

